#include "arccli.h"

#include <iostream>
#include <list>
#include <string>
#include <algorithm>

#include <arc/jobftpcontrol.h>
#include <arc/joblist.h>
#include <arc/mdsquery.h>
#include <arc/notify.h>
#include <arc/resource.h>

#ifdef HAVE_LIBINTL_H
#include <libintl.h>
#define _(A) dgettext("arccli", (A))
#else
#define _(A) (A)
#endif


void arckill (const std::list<std::string> & jobs,
              const std::list<std::string> & clusterselect,
              const std::list<std::string> & clusterreject,
              const std::list<std::string> & status,
              const bool keep,
              const int timeout,
              const bool anonymous) {

	std::list<std::string> jobids =
		GetJobIDsList(jobs, clusterselect, clusterreject);

	if (jobs.empty() && clusterselect.empty() && jobids.empty())
		throw ARCCLIError(_("No jobs"));

	if (jobids.empty())
		throw ARCCLIError(_("No valid jobnames/jobids given"));

	std::list<Job> joblist =
		GetJobInfo(jobids, MDS_FILTER_JOBINFO, anonymous, "", timeout);

	std::list<Queue> queues;
	JobFTPControl ctrl;

	joblist.sort(SortByHost());
	int jobs_killed = 0;
	int jobs_cleaned = 0;
	int jobs_processed = 0;

	for (std::list<Job>::iterator jli = joblist.begin();
	     jli != joblist.end(); jli++) {

		if (jli->status.empty()) {
			notify(WARNING) << _("Job information not found")
			                << ": " << jli->id << std::endl;
			continue;
		}

		if (!status.empty() &&
		    std::find(status.begin(), status.end(), jli->status) == status.end())
			continue;

		if (jli->status == "DELETED") {
			notify(WARNING) << _("Job has already been deleted")
			                << ": " << jli->id << std::endl;
			continue;
		}

		if (jli->status == "FINISHED" || jli->status == "FAILED" ||
		    jli->status == "KILLED") {
			notify(WARNING) << _("Job has already finished")
			                << ": " << jli->id << std::endl;
			continue;
		}

		jobs_processed++;
		notify(INFO) << _("Killing job")
		                << ": " << jli->id << std::endl;

		try {
			ctrl.Cancel(jli->id, timeout, false);
			jobs_killed++;

			if (!keep) {
				notify(INFO) << _("Deleting job")
				                << ": " << jli->id << std::endl;

				ctrl.Clean(jli->id, timeout, false);
				jobs_cleaned++;
				RemoveJobID(jli->id);
			}


			if(INFO <= GetNotifyLevel()) {
				std::string queuestatus = 
				       QueueStatusByJob(*jli, queues, anonymous, timeout);
				if (queuestatus.find("grid-manager is down") == std::string::npos) {
					notify(INFO) << 
					        _("WARNING: Grid-mananger is not functioning "
					          "at the moment. Request to kill/delete the "
					          "job will be processed as soon as it is "
					          "restarted") << std::endl;
				}
			}

		} catch(FTPControlError e) {
			std::string queuestatus;

			if(WARNING <= GetNotifyLevel())
				queuestatus = QueueStatusByJob(*jli, queues, anonymous, timeout);

			if (queuestatus.find("gridftp is down") == std::string::npos) {
				notify(WARNING) << e.what() << std::endl;;
			} else {
				notify(WARNING) << _("Gridftp server of job is down at the "
				                     "moment") << ": " << jli->id << std::endl;
			}
		} catch(ARCLibError& e) {
			notify(WARNING) << e.what() << std::endl;
		}
	}

	notify(WARNING) << _("Jobs processed") << ": " << jobs_processed << ", "
	                << _("killed") << ": " << jobs_killed;
	if(!keep) notify(WARNING) << ", " << _("deleted") << ": " << jobs_cleaned;
	notify(WARNING) << std::endl;
}
